/*
 * gpm2 - mouse driver for the console
 *
 * Copyright (c) 2007         Nico Schottelius <nico-gpm2 () schottelius.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *    daemon specific include file
 ********/

#ifndef GPM2_DAEMON
#define GPM2_DAEMON

/* parameters and options */
#define GPM2_ARGS       "c:f"
#define GPM2_CCONFIG    "/etc/gpm2"

struct gpm2_options {
   char *cconfig;             /* configuration directory */
} opts;

/* functions */
int   commandline(int argc, char **argv);
int   mice_handler();
int   read_config(char *cconfig);
void  set_defaults();
void  usage();


/* structs */

struct mouse {
   /* pointer to init and decode functions */
   int (*open)(int fd);
};

#endif
